#!/usr/bin/env python
# Copyright (C) 2015-2021 Swift Navigation Inc.
# Contact: https://support.swiftnav.com
#
# This source is subject to the license found in the file 'LICENSE' which must
# be be distributed together with this source. All other rights reserved.
#
# THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
# EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.


"""
Inertial Measurement Unit (IMU) messages.
"""

import json

import construct

from sbp.msg import SBP, SENDER_ID
from sbp.utils import fmt_repr, exclude_fields, walk_json_dict, containerize

# Automatically generated from piksi/yaml/swiftnav/sbp/imu.yaml with generate.py.
# Please do not hand edit!


SBP_MSG_IMU_RAW = 0x0900
class MsgImuRaw(SBP):
  """SBP class for message MSG_IMU_RAW (0x0900).

  You can have MSG_IMU_RAW inherit its fields directly
  from an inherited SBP object, or construct it inline using a dict
  of its fields.

  
  Raw data from the Inertial Measurement Unit, containing accelerometer and
  gyroscope readings. The sense of the measurements are to be aligned with the
  indications on the device itself. Measurement units, which are specific to
  the device hardware and settings, are communicated via the MSG_IMU_AUX
  message. If using "time since startup" time tags, the receiving end will
  expect a `MSG_GNSS_TIME_OFFSET` when a PVT fix becomes available to
  synchronise IMU measurements with GNSS. The timestamp must wrap around to
  zero when reaching one week (604800 seconds).

  The time-tagging mode should not change throughout a run.

  Parameters
  ----------
  sbp : SBP
    SBP parent object to inherit from.
  tow : int
    Milliseconds since reference epoch and time status.
  tow_f : int
    Milliseconds since reference epoch, fractional part
  acc_x : int
    Acceleration in the IMU frame X axis
  acc_y : int
    Acceleration in the IMU frame Y axis
  acc_z : int
    Acceleration in the IMU frame Z axis
  gyr_x : int
    Angular rate around IMU frame X axis
  gyr_y : int
    Angular rate around IMU frame Y axis
  gyr_z : int
    Angular rate around IMU frame Z axis
  sender : int
    Optional sender ID, defaults to SENDER_ID (see sbp/msg.py).

  """
  _parser = construct.Struct(
                   'tow' / construct.Int32ul,
                   'tow_f' / construct.Int8ul,
                   'acc_x' / construct.Int16sl,
                   'acc_y' / construct.Int16sl,
                   'acc_z' / construct.Int16sl,
                   'gyr_x' / construct.Int16sl,
                   'gyr_y' / construct.Int16sl,
                   'gyr_z' / construct.Int16sl,)
  __slots__ = [
               'tow',
               'tow_f',
               'acc_x',
               'acc_y',
               'acc_z',
               'gyr_x',
               'gyr_y',
               'gyr_z',
              ]

  def __init__(self, sbp=None, **kwargs):
    if sbp:
      super( MsgImuRaw,
             self).__init__(sbp.msg_type, sbp.sender, sbp.length,
                            sbp.payload, sbp.crc)
      self.from_binary(sbp.payload)
    else:
      super( MsgImuRaw, self).__init__()
      self.msg_type = SBP_MSG_IMU_RAW
      self.sender = kwargs.pop('sender', SENDER_ID)
      self.tow = kwargs.pop('tow')
      self.tow_f = kwargs.pop('tow_f')
      self.acc_x = kwargs.pop('acc_x')
      self.acc_y = kwargs.pop('acc_y')
      self.acc_z = kwargs.pop('acc_z')
      self.gyr_x = kwargs.pop('gyr_x')
      self.gyr_y = kwargs.pop('gyr_y')
      self.gyr_z = kwargs.pop('gyr_z')

  def __repr__(self):
    return fmt_repr(self)

  @staticmethod
  def from_json(s):
    """Given a JSON-encoded string s, build a message object.

    """
    d = json.loads(s)
    return MsgImuRaw.from_json_dict(d)

  @staticmethod
  def from_json_dict(d):
    sbp = SBP.from_json_dict(d)
    return MsgImuRaw(sbp, **d)

 
  def from_binary(self, d):
    """Given a binary payload d, update the appropriate payload fields of
    the message.

    """
    p = MsgImuRaw._parser.parse(d)
    for n in self.__class__.__slots__:
      setattr(self, n, getattr(p, n))

  def to_binary(self):
    """Produce a framed/packed SBP message.

    """
    c = containerize(exclude_fields(self))
    self.payload = MsgImuRaw._parser.build(c)
    return self.pack()

  def into_buffer(self, buf, offset):
    """Produce a framed/packed SBP message into the provided buffer and offset.

    """
    self.payload = containerize(exclude_fields(self))
    self.parser = MsgImuRaw._parser
    self.stream_payload.reset(buf, offset)
    return self.pack_into(buf, offset, self._build_payload)

  def to_json_dict(self):
    self.to_binary()
    d = super( MsgImuRaw, self).to_json_dict()
    j = walk_json_dict(exclude_fields(self))
    d.update(j)
    return d
    
SBP_MSG_IMU_AUX = 0x0901
class MsgImuAux(SBP):
  """SBP class for message MSG_IMU_AUX (0x0901).

  You can have MSG_IMU_AUX inherit its fields directly
  from an inherited SBP object, or construct it inline using a dict
  of its fields.

  
  Auxiliary data specific to a particular IMU. The `imu_type` field will
  always be consistent but the rest of the payload is device specific and
  depends on the value of `imu_type`.

  Parameters
  ----------
  sbp : SBP
    SBP parent object to inherit from.
  imu_type : int
    IMU type
  temp : int
    Raw IMU temperature
  imu_conf : int
    IMU configuration
  sender : int
    Optional sender ID, defaults to SENDER_ID (see sbp/msg.py).

  """
  _parser = construct.Struct(
                   'imu_type' / construct.Int8ul,
                   'temp' / construct.Int16sl,
                   'imu_conf' / construct.Int8ul,)
  __slots__ = [
               'imu_type',
               'temp',
               'imu_conf',
              ]

  def __init__(self, sbp=None, **kwargs):
    if sbp:
      super( MsgImuAux,
             self).__init__(sbp.msg_type, sbp.sender, sbp.length,
                            sbp.payload, sbp.crc)
      self.from_binary(sbp.payload)
    else:
      super( MsgImuAux, self).__init__()
      self.msg_type = SBP_MSG_IMU_AUX
      self.sender = kwargs.pop('sender', SENDER_ID)
      self.imu_type = kwargs.pop('imu_type')
      self.temp = kwargs.pop('temp')
      self.imu_conf = kwargs.pop('imu_conf')

  def __repr__(self):
    return fmt_repr(self)

  @staticmethod
  def from_json(s):
    """Given a JSON-encoded string s, build a message object.

    """
    d = json.loads(s)
    return MsgImuAux.from_json_dict(d)

  @staticmethod
  def from_json_dict(d):
    sbp = SBP.from_json_dict(d)
    return MsgImuAux(sbp, **d)

 
  def from_binary(self, d):
    """Given a binary payload d, update the appropriate payload fields of
    the message.

    """
    p = MsgImuAux._parser.parse(d)
    for n in self.__class__.__slots__:
      setattr(self, n, getattr(p, n))

  def to_binary(self):
    """Produce a framed/packed SBP message.

    """
    c = containerize(exclude_fields(self))
    self.payload = MsgImuAux._parser.build(c)
    return self.pack()

  def into_buffer(self, buf, offset):
    """Produce a framed/packed SBP message into the provided buffer and offset.

    """
    self.payload = containerize(exclude_fields(self))
    self.parser = MsgImuAux._parser
    self.stream_payload.reset(buf, offset)
    return self.pack_into(buf, offset, self._build_payload)

  def to_json_dict(self):
    self.to_binary()
    d = super( MsgImuAux, self).to_json_dict()
    j = walk_json_dict(exclude_fields(self))
    d.update(j)
    return d
    

msg_classes = {
  0x0900: MsgImuRaw,
  0x0901: MsgImuAux,
}


# Copyright (C) 2015-2021 Swift Navigation Inc.
# Contact: https://support.swiftnav.com
#
# This source is subject to the license found in the file 'LICENSE' which must
# be be distributed together with this source. All other rights reserved.
#
# THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
# EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.

from ..msg import crc16, SBP, SBP_PREAMBLE
from ..table import dispatch

import struct
import datetime
import uuid
import six
import warnings


class Framer(six.Iterator):
    """
    Framer

    The :class:`Framer` class frames bytes into SBP messages.
    Implements the Python iterator protocol for received messages.

    Parameters
    ----------
    read : port
      Stream of bytes to read from.
    write : port
      Stream of bytes to write to.
    verbose: boolean
      Increased verbosity of stdout / stderr. Default False
    dispatcher: function
      function to call on each deframed message
    skip_medata: boolean
      do not add "time" and "session-uuid" metadata
    sender_id_filter: list
      list of integer sender_ids to frame, with the exclusion of all others
    """

    def __init__(self,
                 read,
                 write,
                 verbose=False,
                 dispatcher=dispatch,
                 into_buffer=True,
                 skip_metadata=False,
                 sender_id_filter_list=[]):
        self._read = read
        self._write = write
        self._verbose = verbose
        self._broken = False
        self._dispatch = dispatcher
        self._session = str(uuid.uuid4())
        self._buffer = bytearray(16*1024)
        self._into_buffer = into_buffer
        self._skip_metadata = skip_metadata
        self._sender_id_filter_list = sender_id_filter_list

    def __iter__(self):
        self._broken = False
        return self

    def breakiter(self):
        """
        Break out of message iteration.
        May be called from another thread when some thread is blocked in next().
        """
        self._broken = True

    def _time(self):
        """
        Timestamp generator.

        Returns
        -------
        str : ISO 8601 format timestamp
        """
        return datetime.datetime.utcnow().isoformat() + 'Z'

    def __next__(self):
        msg = None
        while msg is None:
            try:
                msg = self._receive()
                if self._broken:
                    raise StopIteration
            except IOError:
                raise StopIteration

        metadata = {}
        if not self._skip_metadata:
            metadata['time'] = self._time()
            metadata['session-uid'] = self._session

        return (msg, metadata)

    def _readall(self, size):
        """
        Read until all bytes are collected.

        Parameters
        ----------
        size : int
          Number of bytes to read.
        """
        data = b""
        while len(data) < size:
            d = self._read(size - len(data))
            if not d or self._broken:
                raise StopIteration
            data += d
        return data

    def _receive(self):
        """
        Read and build SBP message.
        """
        # preamble - not readall(1) to allow breaking before messages,
        # empty input
        preamble = self._read(1)
        if not preamble:
            return None
        elif ord(preamble) != SBP_PREAMBLE:
            if self._verbose:
                print("Host Side Unhandled byte: 0x%02x" % ord(preamble))
            return None
        # hdr
        hdr = self._readall(5)
        msg_crc = crc16(hdr)
        msg_type, sender, msg_len = struct.unpack("<HHB", hdr)
        # data
        data = self._readall(msg_len)
        msg_crc = crc16(data, msg_crc)
        # crc
        crc = self._readall(2)
        crc, = struct.unpack("<H", crc)
        if crc != msg_crc:
            if self._verbose:
                print("crc mismatch: 0x%04X 0x%04X" % (msg_crc, crc))
            return None
        if (len(self._sender_id_filter_list) == 0 or sender in self._sender_id_filter_list):
            msg = SBP(msg_type, sender, msg_len, data, crc)
            try:
                msg = self._dispatch(msg)
            except Exception as exc:
                warnings.warn("SBP dispatch error: %s" % (exc,))
            return msg
        else:
            return None

    def __call__(self, *msgs, **metadata):
        """
        Build and write SBP message.

        Parameters
        ----------
        msgs : SBP messages
          SBP messages to send.  Multiple messages are sent in one batch.
        metadata : dict
          Metadata for this batch of messages, e.g. `{'time': 'ISO 8601 str'}`
          (ignored for now).
        """
        index = 0
        if self._into_buffer:
            for msg in msgs:
                index += msg.into_buffer(self._buffer, index)
        else:
            for msg in msgs:
                msg_buff = msg.to_binary()
                buff_len = len(msg_buff)
                self._buffer[index:(index+buff_len)] = bytearray(msg_buff)
                index += buff_len
        self._write(memoryview(self._buffer)[:index])



# Copyright (C) 2015-2021 Swift Navigation Inc.
# Contact: https://support.swiftnav.com
#
# This source is subject to the license found in the file 'LICENSE' which must
# be be distributed together with this source. All other rights reserved.
#
# THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
# EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.

from .base_driver import BaseDriver

import datetime
import json
import time


class FileDriver(BaseDriver):
    """
    FileDriver

    The :class:`FileDriver` class wraps file handle IO

    Parameters
    ----------
    fd : File descriptor
      File stream.
    """

    def _read(self, size):
        """
        Read wrapper.

        Parameters
        ----------
        size : int
          Number of bytes to read.
        """
        return_val = self.handle.read(size)
        if not return_val:
            raise IOError
        else:
            return return_val

    def __init__(self, fd):
        super(FileDriver, self).__init__(fd)

    def readline(self):
        return self.handle.readline()


class PlaybackFileDriver(FileDriver):
    """
    PlaybackFileDriver

    The :class:`PlaybackFileDriver` class for SBP log playback. Main purpose is to delay the
    messages to emulate real data.

    Parameters
    ----------
    fd : File handle
      File containing the playback data.
    """

    def __init__(self, fd):
        super(PlaybackFileDriver, self).__init__(fd)
        self._start_datetime = None
        self._start_timestamp = None

    def _delay(self, line):
        """
        Creates delay between messages based on timestamp metadata. Method records the first
        available timestamp and the real datetime of handling that timestamp. This enables two
        separate timelines; real and play(back). Subsequent calls try to keep these timelines insync
        as tightly as possible by utilising sleep whenever necessary.

        Parameters
        ----------
        size : int
          Number of bytes to read.
        """
        data = json.loads(line)

        if not "time" in data:
            return

        timestamp = datetime.datetime.strptime(data["time"], "%Y-%m-%dT%H:%M:%S.%fZ")

        if self._start_datetime and self._start_timestamp:
            real = datetime.datetime.now() - self._start_datetime
            play = timestamp - self._start_timestamp
            delay = (play - real).total_seconds()

            if delay > 0:
                time.sleep(delay)
        else:
            self._start_datetime = datetime.datetime.now()
            self._start_timestamp = timestamp

    def readline(self):
        """
        Readline wrapper. Delays the return as necessary based on timestamps.
        """
        line = super(PlaybackFileDriver, self).readline()

        self._delay(line)

        return line



# Copyright (C) 2017-2021 Swift Navigation Inc.
# Contact: https://support.swiftnav.com
#
# This source is subject to the license found in the file 'LICENSE' which must
# be be distributed together with this source. All other rights reserved.
#
# THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
# EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.
"""
the :mod:`sbp.client.util.settingmonitor` module contains functionality to
monitor SBP_MSG_SETTINGS_READ_RESP messages for desired values
"""

from sbp.settings import SBP_MSG_SETTINGS_READ_RESP

import time


class SettingMonitor(object):
    """Class to monitor Settings via SBP messages

    Parameters
    ----------
    None

    """

    def __init__(self):
        self.settings = []

    def capture_setting(self, sbp_msg, **metadata):
        """Callback to extract and store setting values from
        SBP_MSG_SETTINGS_READ_RESP

        Messages of any type other than SBP_MSG_SETTINGS_READ_RESP are ignored
        """
        if sbp_msg.msg_type == SBP_MSG_SETTINGS_READ_RESP:
            section, setting, value = sbp_msg.payload.split(b'\0')[:3]
            self.settings.append((section, setting, value))

    def wait_for_setting_value(self, section, setting, value, wait_time=5.0):
        """Function to wait wait_time seconds to see a
        SBP_MSG_SETTINGS_READ_RESP message with a user-specified value
        """
        expire = time.time() + wait_time
        ok = False
        while not ok and time.time() < expire:
            settings = [x for x in self.settings if (x[0], x[1]) == (section, setting)]
            # Check to see if the last setting has the value we want
            if len(settings) > 0:
                ok = settings[-1][2] == value

            time.sleep(0.1)
        return ok

    def clear(self, section=None, setting=None, value=None):
        """Clear settings"""
        match = [all((section is None or x_y_z[0] == section,
                      setting is None or x_y_z[1] == setting,
                      value is None or x_y_z[2] == value)) for x_y_z in self.settings]

        keep = [setting_remove for setting_remove in zip(self.settings,match) if not setting_remove[1]]

        self.settings[:] = [x[0] for x in keep]


if __name__ == "__main__":
    from sbp.client import Handler, Framer
    from sbp.client.drivers.network_drivers import TCPDriver
    from sbp.settings import MsgSettingsWrite, MsgSettingsReadReq
    import argparse

    def print_setting(sbp_msg, **metadata):
        print(sbp_msg)

    def main():
        parser = argparse.ArgumentParser(
            description="Swift Navigation SBP Setting Monitor example.")
        parser.add_argument(
            "-H",
            "--host",
            required=True,
            help="specify the host address.")
        parser.add_argument(
            "-p",
            "--port",
            default=55556,
            help="specify the port to use.")
        args = parser.parse_args()

        monitor = SettingMonitor()

        with TCPDriver(args.host, args.port) as driver:
            with Handler(Framer(driver.read, driver.write, verbose=True)) as link:
                driver.flush()
                time.sleep(2)

                # Capture setting messages
                link.add_callback(monitor.capture_setting,
                                  SBP_MSG_SETTINGS_READ_RESP)

                link.add_callback(print_setting, SBP_MSG_SETTINGS_READ_RESP)

                # Disable spectrum analyzer
                link(MsgSettingsWrite(setting=b'%s\0%s\0%s\0' % (
                    b'system_monitor', b'spectrum_analyzer', b'False')))

                # Request the value of the system_monitor:spectrum_analyzer setting
                link(MsgSettingsReadReq(setting=b'%s\0%s\0' % (
                     b'system_monitor', b'spectrum_analyzer')))

                # Wait up to 5 seconds to see the setting we want
                specan_off = monitor.wait_for_setting_value(
                        b'system_monitor', b'spectrum_analyzer', b'False')

                assert(specan_off == True)
                print("Spectrum analyzer turned off!")

                # Request the value of the system_monitor:spectrum_analyzer setting
                link(MsgSettingsReadReq(setting=b'%s\0%s\0' % (
                     b'system_monitor', b'spectrum_analyzer')))

                # Wait up to 5 seconds to see the setting we (don't) want
                specan_off = monitor.wait_for_setting_value(
                        b'system_monitor', b'spectrum_analyzer', b'True')

                assert(specan_off == False)
                print("Spectrum analyzer still off!")

    main()



#!/usr/bin/env python
# Copyright (C) 2015-2021 Swift Navigation Inc.
# Contact: https://support.swiftnav.com
#
# This source is subject to the license found in the file 'LICENSE' which must
# be be distributed together with this source. All other rights reserved.
#
# THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
# EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.


"""
Messages for reading, writing, and discovering device settings. Settings
with a "string" field have multiple values in this field delimited with a
null character (the c style null terminator).  For instance, when querying
the 'firmware_version' setting in the 'system_info' section, the following
array of characters needs to be sent for the string field in
MSG_SETTINGS_READ: "system_info\0firmware_version\0", where the delimiting
null characters are specified with the escape sequence '\0' and all
quotation marks should be omitted.

In the message descriptions below, the generic strings SECTION_SETTING and
SETTING are used to refer to the two strings that comprise the identifier of
an individual setting.In firmware_version example above, SECTION_SETTING is
the 'system_info', and the SETTING portion is 'firmware_version'.
See the "Software Settings Manual" on support.swiftnav.com for detailed
documentation about all settings and sections available for each Swift
firmware version. Settings manuals are available for each firmware version
at the following link:
https://support.swiftnav.com/support/solutions/articles/44001850753-piksi-multi-specification.
The latest settings document is also available at the following link:
http://swiftnav.com/latest/piksi-multi-settings . See lastly
https://github.com/swift-nav/piksi_tools/blob/master/piksi_tools/settings.py
, the open source python command line utility for reading, writing, and
saving settings in the piksi_tools repository on github as a helpful
reference and example.
"""

import json

import construct

from sbp.msg import SBP, SENDER_ID
from sbp.utils import fmt_repr, exclude_fields, walk_json_dict, containerize

# Automatically generated from piksi/yaml/swiftnav/sbp/settings.yaml with generate.py.
# Please do not hand edit!


SBP_MSG_SETTINGS_SAVE = 0x00A1
class MsgSettingsSave(SBP):
  """SBP class for message MSG_SETTINGS_SAVE (0x00A1).

  You can have MSG_SETTINGS_SAVE inherit its fields directly
  from an inherited SBP object, or construct it inline using a dict
  of its fields.

  
  The save settings message persists the device's current settings
  configuration to its onboard flash memory file system.

  """
  __slots__ = []

  def __init__(self, sbp=None, **kwargs):
    if sbp:
      super( MsgSettingsSave,
             self).__init__(sbp.msg_type, sbp.sender, sbp.length,
                            sbp.payload, sbp.crc)
      self.payload = sbp.payload
    else:
      super( MsgSettingsSave, self).__init__()
      self.msg_type = SBP_MSG_SETTINGS_SAVE
      self.sender = kwargs.pop('sender', SENDER_ID)
      self.payload = b""

  def __repr__(self):
    return fmt_repr(self)

  @staticmethod
  def from_json(s):
    """Given a JSON-encoded string s, build a message object.

    """
    d = json.loads(s)
    return MsgSettingsSave.from_json_dict(d)

  @staticmethod
  def from_json_dict(d):
    sbp = SBP.from_json_dict(d)
    return MsgSettingsSave(sbp, **d)

 
    
SBP_MSG_SETTINGS_WRITE = 0x00A0
class MsgSettingsWrite(SBP):
  r"""SBP class for message MSG_SETTINGS_WRITE (0x00A0).

  You can have MSG_SETTINGS_WRITE inherit its fields directly
  from an inherited SBP object, or construct it inline using a dict
  of its fields.

  
  The setting message writes the device configuration for a particular setting
  via A NULL-terminated and NULL-delimited string with contents
  "SECTION_SETTING\0SETTING\0VALUE\0" where the '\0' escape sequence denotes
  the NULL character and where quotation marks are omitted. A device will only
  process to this message when it is received from sender ID 0x42. An example
  string that could be sent to a device is "solution\0soln_freq\010\0".

  Parameters
  ----------
  sbp : SBP
    SBP parent object to inherit from.
  setting : string
    A NULL-terminated and NULL-delimited string with contents
    "SECTION_SETTING\0SETTING\0VALUE\0"
  sender : int
    Optional sender ID, defaults to SENDER_ID (see sbp/msg.py).

  """
  _parser = construct.Struct(
                   'setting' / construct.GreedyBytes,)
  __slots__ = [
               'setting',
              ]

  def __init__(self, sbp=None, **kwargs):
    if sbp:
      super( MsgSettingsWrite,
             self).__init__(sbp.msg_type, sbp.sender, sbp.length,
                            sbp.payload, sbp.crc)
      self.from_binary(sbp.payload)
    else:
      super( MsgSettingsWrite, self).__init__()
      self.msg_type = SBP_MSG_SETTINGS_WRITE
      self.sender = kwargs.pop('sender', SENDER_ID)
      self.setting = kwargs.pop('setting')

  def __repr__(self):
    return fmt_repr(self)

  @staticmethod
  def from_json(s):
    """Given a JSON-encoded string s, build a message object.

    """
    d = json.loads(s)
    return MsgSettingsWrite.from_json_dict(d)

  @staticmethod
  def from_json_dict(d):
    sbp = SBP.from_json_dict(d)
    return MsgSettingsWrite(sbp, **d)

 
  def from_binary(self, d):
    """Given a binary payload d, update the appropriate payload fields of
    the message.

    """
    p = MsgSettingsWrite._parser.parse(d)
    for n in self.__class__.__slots__:
      setattr(self, n, getattr(p, n))

  def to_binary(self):
    """Produce a framed/packed SBP message.

    """
    c = containerize(exclude_fields(self))
    self.payload = MsgSettingsWrite._parser.build(c)
    return self.pack()

  def into_buffer(self, buf, offset):
    """Produce a framed/packed SBP message into the provided buffer and offset.

    """
    self.payload = containerize(exclude_fields(self))
    self.parser = MsgSettingsWrite._parser
    self.stream_payload.reset(buf, offset)
    return self.pack_into(buf, offset, self._build_payload)

  def to_json_dict(self):
    self.to_binary()
    d = super( MsgSettingsWrite, self).to_json_dict()
    j = walk_json_dict(exclude_fields(self))
    d.update(j)
    return d
    
SBP_MSG_SETTINGS_WRITE_RESP = 0x00AF
class MsgSettingsWriteResp(SBP):
  r"""SBP class for message MSG_SETTINGS_WRITE_RESP (0x00AF).

  You can have MSG_SETTINGS_WRITE_RESP inherit its fields directly
  from an inherited SBP object, or construct it inline using a dict
  of its fields.

  
  Return the status of a write request with the new value of the setting.  If
  the requested value is rejected, the current value will be returned. The
  string field is a NULL-terminated and NULL-delimited string with contents
  "SECTION_SETTING\0SETTING\0VALUE\0" where the '\0' escape sequence denotes
  the NULL character and where quotation marks are omitted. An example string
  that could be sent from device is "solution\0soln_freq\010\0".

  Parameters
  ----------
  sbp : SBP
    SBP parent object to inherit from.
  status : int
    Write status
  setting : string
    A NULL-terminated and delimited string with contents
    "SECTION_SETTING\0SETTING\0VALUE\0"
  sender : int
    Optional sender ID, defaults to SENDER_ID (see sbp/msg.py).

  """
  _parser = construct.Struct(
                   'status' / construct.Int8ul,
                   'setting' / construct.GreedyBytes,)
  __slots__ = [
               'status',
               'setting',
              ]

  def __init__(self, sbp=None, **kwargs):
    if sbp:
      super( MsgSettingsWriteResp,
             self).__init__(sbp.msg_type, sbp.sender, sbp.length,
                            sbp.payload, sbp.crc)
      self.from_binary(sbp.payload)
    else:
      super( MsgSettingsWriteResp, self).__init__()
      self.msg_type = SBP_MSG_SETTINGS_WRITE_RESP
      self.sender = kwargs.pop('sender', SENDER_ID)
      self.status = kwargs.pop('status')
      self.setting = kwargs.pop('setting')

  def __repr__(self):
    return fmt_repr(self)

  @staticmethod
  def from_json(s):
    """Given a JSON-encoded string s, build a message object.

    """
    d = json.loads(s)
    return MsgSettingsWriteResp.from_json_dict(d)

  @staticmethod
  def from_json_dict(d):
    sbp = SBP.from_json_dict(d)
    return MsgSettingsWriteResp(sbp, **d)

 
  def from_binary(self, d):
    """Given a binary payload d, update the appropriate payload fields of
    the message.

    """
    p = MsgSettingsWriteResp._parser.parse(d)
    for n in self.__class__.__slots__:
      setattr(self, n, getattr(p, n))

  def to_binary(self):
    """Produce a framed/packed SBP message.

    """
    c = containerize(exclude_fields(self))
    self.payload = MsgSettingsWriteResp._parser.build(c)
    return self.pack()

  def into_buffer(self, buf, offset):
    """Produce a framed/packed SBP message into the provided buffer and offset.

    """
    self.payload = containerize(exclude_fields(self))
    self.parser = MsgSettingsWriteResp._parser
    self.stream_payload.reset(buf, offset)
    return self.pack_into(buf, offset, self._build_payload)

  def to_json_dict(self):
    self.to_binary()
    d = super( MsgSettingsWriteResp, self).to_json_dict()
    j = walk_json_dict(exclude_fields(self))
    d.update(j)
    return d
    
SBP_MSG_SETTINGS_READ_REQ = 0x00A4
class MsgSettingsReadReq(SBP):
  r"""SBP class for message MSG_SETTINGS_READ_REQ (0x00A4).

  You can have MSG_SETTINGS_READ_REQ inherit its fields directly
  from an inherited SBP object, or construct it inline using a dict
  of its fields.

  
  The setting message that reads the device configuration. The string field is
  a NULL-terminated and NULL-delimited string with contents
  "SECTION_SETTING\0SETTING\0" where the '\0' escape sequence denotes the NULL
  character and where quotation marks are omitted. An example string that
  could be sent to a device is "solution\0soln_freq\0". A device will only
  respond to this message when it is received from sender ID 0x42. A device
  should respond with a MSG_SETTINGS_READ_RESP message (msg_id 0x00A5).

  Parameters
  ----------
  sbp : SBP
    SBP parent object to inherit from.
  setting : string
    A NULL-terminated and NULL-delimited string with contents
    "SECTION_SETTING\0SETTING\0"
  sender : int
    Optional sender ID, defaults to SENDER_ID (see sbp/msg.py).

  """
  _parser = construct.Struct(
                   'setting' / construct.GreedyBytes,)
  __slots__ = [
               'setting',
              ]

  def __init__(self, sbp=None, **kwargs):
    if sbp:
      super( MsgSettingsReadReq,
             self).__init__(sbp.msg_type, sbp.sender, sbp.length,
                            sbp.payload, sbp.crc)
      self.from_binary(sbp.payload)
    else:
      super( MsgSettingsReadReq, self).__init__()
      self.msg_type = SBP_MSG_SETTINGS_READ_REQ
      self.sender = kwargs.pop('sender', SENDER_ID)
      self.setting = kwargs.pop('setting')

  def __repr__(self):
    return fmt_repr(self)

  @staticmethod
  def from_json(s):
    """Given a JSON-encoded string s, build a message object.

    """
    d = json.loads(s)
    return MsgSettingsReadReq.from_json_dict(d)

  @staticmethod
  def from_json_dict(d):
    sbp = SBP.from_json_dict(d)
    return MsgSettingsReadReq(sbp, **d)

 
  def from_binary(self, d):
    """Given a binary payload d, update the appropriate payload fields of
    the message.

    """
    p = MsgSettingsReadReq._parser.parse(d)
    for n in self.__class__.__slots__:
      setattr(self, n, getattr(p, n))

  def to_binary(self):
    """Produce a framed/packed SBP message.

    """
    c = containerize(exclude_fields(self))
    self.payload = MsgSettingsReadReq._parser.build(c)
    return self.pack()

  def into_buffer(self, buf, offset):
    """Produce a framed/packed SBP message into the provided buffer and offset.

    """
    self.payload = containerize(exclude_fields(self))
    self.parser = MsgSettingsReadReq._parser
    self.stream_payload.reset(buf, offset)
    return self.pack_into(buf, offset, self._build_payload)

  def to_json_dict(self):
    self.to_binary()
    d = super( MsgSettingsReadReq, self).to_json_dict()
    j = walk_json_dict(exclude_fields(self))
    d.update(j)
    return d
    
SBP_MSG_SETTINGS_READ_RESP = 0x00A5
class MsgSettingsReadResp(SBP):
  r"""SBP class for message MSG_SETTINGS_READ_RESP (0x00A5).

  You can have MSG_SETTINGS_READ_RESP inherit its fields directly
  from an inherited SBP object, or construct it inline using a dict
  of its fields.

  
  The setting message with which the device responds after a
  MSG_SETTING_READ_REQ is sent to device. The string field is a NULL-
  terminated and NULL-delimited string with contents
  "SECTION_SETTING\0SETTING\0VALUE\0" where the '\0' escape sequence denotes
  the NULL character and where quotation marks are omitted. An example string
  that could be sent from device is "solution\0soln_freq\010\0".

  Parameters
  ----------
  sbp : SBP
    SBP parent object to inherit from.
  setting : string
    A NULL-terminated and NULL-delimited string with contents
    "SECTION_SETTING\0SETTING\0VALUE\0"
  sender : int
    Optional sender ID, defaults to SENDER_ID (see sbp/msg.py).

  """
  _parser = construct.Struct(
                   'setting' / construct.GreedyBytes,)
  __slots__ = [
               'setting',
              ]

  def __init__(self, sbp=None, **kwargs):
    if sbp:
      super( MsgSettingsReadResp,
             self).__init__(sbp.msg_type, sbp.sender, sbp.length,
                            sbp.payload, sbp.crc)
      self.from_binary(sbp.payload)
    else:
      super( MsgSettingsReadResp, self).__init__()
      self.msg_type = SBP_MSG_SETTINGS_READ_RESP
      self.sender = kwargs.pop('sender', SENDER_ID)
      self.setting = kwargs.pop('setting')

  def __repr__(self):
    return fmt_repr(self)

  @staticmethod
  def from_json(s):
    """Given a JSON-encoded string s, build a message object.

    """
    d = json.loads(s)
    return MsgSettingsReadResp.from_json_dict(d)

  @staticmethod
  def from_json_dict(d):
    sbp = SBP.from_json_dict(d)
    return MsgSettingsReadResp(sbp, **d)

 
  def from_binary(self, d):
    """Given a binary payload d, update the appropriate payload fields of
    the message.

    """
    p = MsgSettingsReadResp._parser.parse(d)
    for n in self.__class__.__slots__:
      setattr(self, n, getattr(p, n))

  def to_binary(self):
    """Produce a framed/packed SBP message.

    """
    c = containerize(exclude_fields(self))
    self.payload = MsgSettingsReadResp._parser.build(c)
    return self.pack()

  def into_buffer(self, buf, offset):
    """Produce a framed/packed SBP message into the provided buffer and offset.

    """
    self.payload = containerize(exclude_fields(self))
    self.parser = MsgSettingsReadResp._parser
    self.stream_payload.reset(buf, offset)
    return self.pack_into(buf, offset, self._build_payload)

  def to_json_dict(self):
    self.to_binary()
    d = super( MsgSettingsReadResp, self).to_json_dict()
    j = walk_json_dict(exclude_fields(self))
    d.update(j)
    return d
    
SBP_MSG_SETTINGS_READ_BY_INDEX_REQ = 0x00A2
class MsgSettingsReadByIndexReq(SBP):
  """SBP class for message MSG_SETTINGS_READ_BY_INDEX_REQ (0x00A2).

  You can have MSG_SETTINGS_READ_BY_INDEX_REQ inherit its fields directly
  from an inherited SBP object, or construct it inline using a dict
  of its fields.

  
  The settings message for iterating through the settings values. A device
  will respond to this message with a "MSG_SETTINGS_READ_BY_INDEX_RESP".

  Parameters
  ----------
  sbp : SBP
    SBP parent object to inherit from.
  index : int
    An index into the device settings, with values ranging from 0 to
    length(settings).
  sender : int
    Optional sender ID, defaults to SENDER_ID (see sbp/msg.py).

  """
  _parser = construct.Struct(
                   'index' / construct.Int16ul,)
  __slots__ = [
               'index',
              ]

  def __init__(self, sbp=None, **kwargs):
    if sbp:
      super( MsgSettingsReadByIndexReq,
             self).__init__(sbp.msg_type, sbp.sender, sbp.length,
                            sbp.payload, sbp.crc)
      self.from_binary(sbp.payload)
    else:
      super( MsgSettingsReadByIndexReq, self).__init__()
      self.msg_type = SBP_MSG_SETTINGS_READ_BY_INDEX_REQ
      self.sender = kwargs.pop('sender', SENDER_ID)
      self.index = kwargs.pop('index')

  def __repr__(self):
    return fmt_repr(self)

  @staticmethod
  def from_json(s):
    """Given a JSON-encoded string s, build a message object.

    """
    d = json.loads(s)
    return MsgSettingsReadByIndexReq.from_json_dict(d)

  @staticmethod
  def from_json_dict(d):
    sbp = SBP.from_json_dict(d)
    return MsgSettingsReadByIndexReq(sbp, **d)

 
  def from_binary(self, d):
    """Given a binary payload d, update the appropriate payload fields of
    the message.

    """
    p = MsgSettingsReadByIndexReq._parser.parse(d)
    for n in self.__class__.__slots__:
      setattr(self, n, getattr(p, n))

  def to_binary(self):
    """Produce a framed/packed SBP message.

    """
    c = containerize(exclude_fields(self))
    self.payload = MsgSettingsReadByIndexReq._parser.build(c)
    return self.pack()

  def into_buffer(self, buf, offset):
    """Produce a framed/packed SBP message into the provided buffer and offset.

    """
    self.payload = containerize(exclude_fields(self))
    self.parser = MsgSettingsReadByIndexReq._parser
    self.stream_payload.reset(buf, offset)
    return self.pack_into(buf, offset, self._build_payload)

  def to_json_dict(self):
    self.to_binary()
    d = super( MsgSettingsReadByIndexReq, self).to_json_dict()
    j = walk_json_dict(exclude_fields(self))
    d.update(j)
    return d
    
SBP_MSG_SETTINGS_READ_BY_INDEX_RESP = 0x00A7
class MsgSettingsReadByIndexResp(SBP):
  r"""SBP class for message MSG_SETTINGS_READ_BY_INDEX_RESP (0x00A7).

  You can have MSG_SETTINGS_READ_BY_INDEX_RESP inherit its fields directly
  from an inherited SBP object, or construct it inline using a dict
  of its fields.

  
  The settings message that reports the value of a setting at an index.

  In the string field, it reports NULL-terminated and delimited string with
  contents "SECTION_SETTING\0SETTING\0VALUE\0FORMAT_TYPE\0". where the '\0'
  escape sequence denotes the NULL character and where quotation marks are
  omitted. The FORMAT_TYPE field is optional and denotes possible string
  values of the setting as a hint to the user. If included, the format type
  portion of the string has the format "enum:value1,value2,value3". An example
  string that could be sent from the device is
  "simulator\0enabled\0True\0enum:True,False\0".

  Parameters
  ----------
  sbp : SBP
    SBP parent object to inherit from.
  index : int
    An index into the device settings, with values ranging from 0 to
    length(settings)
  setting : string
    A NULL-terminated and delimited string with contents
    "SECTION_SETTING\0SETTING\0VALUE\0FORMAT_TYPE\0"
  sender : int
    Optional sender ID, defaults to SENDER_ID (see sbp/msg.py).

  """
  _parser = construct.Struct(
                   'index' / construct.Int16ul,
                   'setting' / construct.GreedyBytes,)
  __slots__ = [
               'index',
               'setting',
              ]

  def __init__(self, sbp=None, **kwargs):
    if sbp:
      super( MsgSettingsReadByIndexResp,
             self).__init__(sbp.msg_type, sbp.sender, sbp.length,
                            sbp.payload, sbp.crc)
      self.from_binary(sbp.payload)
    else:
      super( MsgSettingsReadByIndexResp, self).__init__()
      self.msg_type = SBP_MSG_SETTINGS_READ_BY_INDEX_RESP
      self.sender = kwargs.pop('sender', SENDER_ID)
      self.index = kwargs.pop('index')
      self.setting = kwargs.pop('setting')

  def __repr__(self):
    return fmt_repr(self)

  @staticmethod
  def from_json(s):
    """Given a JSON-encoded string s, build a message object.

    """
    d = json.loads(s)
    return MsgSettingsReadByIndexResp.from_json_dict(d)

  @staticmethod
  def from_json_dict(d):
    sbp = SBP.from_json_dict(d)
    return MsgSettingsReadByIndexResp(sbp, **d)

 
  def from_binary(self, d):
    """Given a binary payload d, update the appropriate payload fields of
    the message.

    """
    p = MsgSettingsReadByIndexResp._parser.parse(d)
    for n in self.__class__.__slots__:
      setattr(self, n, getattr(p, n))

  def to_binary(self):
    """Produce a framed/packed SBP message.

    """
    c = containerize(exclude_fields(self))
    self.payload = MsgSettingsReadByIndexResp._parser.build(c)
    return self.pack()

  def into_buffer(self, buf, offset):
    """Produce a framed/packed SBP message into the provided buffer and offset.

    """
    self.payload = containerize(exclude_fields(self))
    self.parser = MsgSettingsReadByIndexResp._parser
    self.stream_payload.reset(buf, offset)
    return self.pack_into(buf, offset, self._build_payload)

  def to_json_dict(self):
    self.to_binary()
    d = super( MsgSettingsReadByIndexResp, self).to_json_dict()
    j = walk_json_dict(exclude_fields(self))
    d.update(j)
    return d
    
SBP_MSG_SETTINGS_READ_BY_INDEX_DONE = 0x00A6
class MsgSettingsReadByIndexDone(SBP):
  """SBP class for message MSG_SETTINGS_READ_BY_INDEX_DONE (0x00A6).

  You can have MSG_SETTINGS_READ_BY_INDEX_DONE inherit its fields directly
  from an inherited SBP object, or construct it inline using a dict
  of its fields.

  
  The settings message for indicating end of the settings values.

  """
  __slots__ = []

  def __init__(self, sbp=None, **kwargs):
    if sbp:
      super( MsgSettingsReadByIndexDone,
             self).__init__(sbp.msg_type, sbp.sender, sbp.length,
                            sbp.payload, sbp.crc)
      self.payload = sbp.payload
    else:
      super( MsgSettingsReadByIndexDone, self).__init__()
      self.msg_type = SBP_MSG_SETTINGS_READ_BY_INDEX_DONE
      self.sender = kwargs.pop('sender', SENDER_ID)
      self.payload = b""

  def __repr__(self):
    return fmt_repr(self)

  @staticmethod
  def from_json(s):
    """Given a JSON-encoded string s, build a message object.

    """
    d = json.loads(s)
    return MsgSettingsReadByIndexDone.from_json_dict(d)

  @staticmethod
  def from_json_dict(d):
    sbp = SBP.from_json_dict(d)
    return MsgSettingsReadByIndexDone(sbp, **d)

 
    
SBP_MSG_SETTINGS_REGISTER = 0x00AE
class MsgSettingsRegister(SBP):
  r"""SBP class for message MSG_SETTINGS_REGISTER (0x00AE).

  You can have MSG_SETTINGS_REGISTER inherit its fields directly
  from an inherited SBP object, or construct it inline using a dict
  of its fields.

  
  This message registers the presence and default value of a setting with a
  settings daemon.  The host should reply with MSG_SETTINGS_WRITE for this
  setting to set the initial value.

  Parameters
  ----------
  sbp : SBP
    SBP parent object to inherit from.
  setting : string
    A NULL-terminated and delimited string with contents
    "SECTION_SETTING\0SETTING\0VALUE".
  sender : int
    Optional sender ID, defaults to SENDER_ID (see sbp/msg.py).

  """
  _parser = construct.Struct(
                   'setting' / construct.GreedyBytes,)
  __slots__ = [
               'setting',
              ]

  def __init__(self, sbp=None, **kwargs):
    if sbp:
      super( MsgSettingsRegister,
             self).__init__(sbp.msg_type, sbp.sender, sbp.length,
                            sbp.payload, sbp.crc)
      self.from_binary(sbp.payload)
    else:
      super( MsgSettingsRegister, self).__init__()
      self.msg_type = SBP_MSG_SETTINGS_REGISTER
      self.sender = kwargs.pop('sender', SENDER_ID)
      self.setting = kwargs.pop('setting')

  def __repr__(self):
    return fmt_repr(self)

  @staticmethod
  def from_json(s):
    """Given a JSON-encoded string s, build a message object.

    """
    d = json.loads(s)
    return MsgSettingsRegister.from_json_dict(d)

  @staticmethod
  def from_json_dict(d):
    sbp = SBP.from_json_dict(d)
    return MsgSettingsRegister(sbp, **d)

 
  def from_binary(self, d):
    """Given a binary payload d, update the appropriate payload fields of
    the message.

    """
    p = MsgSettingsRegister._parser.parse(d)
    for n in self.__class__.__slots__:
      setattr(self, n, getattr(p, n))

  def to_binary(self):
    """Produce a framed/packed SBP message.

    """
    c = containerize(exclude_fields(self))
    self.payload = MsgSettingsRegister._parser.build(c)
    return self.pack()

  def into_buffer(self, buf, offset):
    """Produce a framed/packed SBP message into the provided buffer and offset.

    """
    self.payload = containerize(exclude_fields(self))
    self.parser = MsgSettingsRegister._parser
    self.stream_payload.reset(buf, offset)
    return self.pack_into(buf, offset, self._build_payload)

  def to_json_dict(self):
    self.to_binary()
    d = super( MsgSettingsRegister, self).to_json_dict()
    j = walk_json_dict(exclude_fields(self))
    d.update(j)
    return d
    
SBP_MSG_SETTINGS_REGISTER_RESP = 0x01AF
class MsgSettingsRegisterResp(SBP):
  r"""SBP class for message MSG_SETTINGS_REGISTER_RESP (0x01AF).

  You can have MSG_SETTINGS_REGISTER_RESP inherit its fields directly
  from an inherited SBP object, or construct it inline using a dict
  of its fields.

  
  This message responds to setting registration with the effective value. The
  effective value shall differ from the given default value if setting was
  already registered or is available in the permanent setting storage and had
  a different value.

  Parameters
  ----------
  sbp : SBP
    SBP parent object to inherit from.
  status : int
    Register status
  setting : string
    A NULL-terminated and delimited string with contents
    "SECTION_SETTING\0SETTING\0VALUE". The meaning of value is defined
    according to the status field.
  sender : int
    Optional sender ID, defaults to SENDER_ID (see sbp/msg.py).

  """
  _parser = construct.Struct(
                   'status' / construct.Int8ul,
                   'setting' / construct.GreedyBytes,)
  __slots__ = [
               'status',
               'setting',
              ]

  def __init__(self, sbp=None, **kwargs):
    if sbp:
      super( MsgSettingsRegisterResp,
             self).__init__(sbp.msg_type, sbp.sender, sbp.length,
                            sbp.payload, sbp.crc)
      self.from_binary(sbp.payload)
    else:
      super( MsgSettingsRegisterResp, self).__init__()
      self.msg_type = SBP_MSG_SETTINGS_REGISTER_RESP
      self.sender = kwargs.pop('sender', SENDER_ID)
      self.status = kwargs.pop('status')
      self.setting = kwargs.pop('setting')

  def __repr__(self):
    return fmt_repr(self)

  @staticmethod
  def from_json(s):
    """Given a JSON-encoded string s, build a message object.

    """
    d = json.loads(s)
    return MsgSettingsRegisterResp.from_json_dict(d)

  @staticmethod
  def from_json_dict(d):
    sbp = SBP.from_json_dict(d)
    return MsgSettingsRegisterResp(sbp, **d)

 
  def from_binary(self, d):
    """Given a binary payload d, update the appropriate payload fields of
    the message.

    """
    p = MsgSettingsRegisterResp._parser.parse(d)
    for n in self.__class__.__slots__:
      setattr(self, n, getattr(p, n))

  def to_binary(self):
    """Produce a framed/packed SBP message.

    """
    c = containerize(exclude_fields(self))
    self.payload = MsgSettingsRegisterResp._parser.build(c)
    return self.pack()

  def into_buffer(self, buf, offset):
    """Produce a framed/packed SBP message into the provided buffer and offset.

    """
    self.payload = containerize(exclude_fields(self))
    self.parser = MsgSettingsRegisterResp._parser
    self.stream_payload.reset(buf, offset)
    return self.pack_into(buf, offset, self._build_payload)

  def to_json_dict(self):
    self.to_binary()
    d = super( MsgSettingsRegisterResp, self).to_json_dict()
    j = walk_json_dict(exclude_fields(self))
    d.update(j)
    return d
    

msg_classes = {
  0x00A1: MsgSettingsSave,
  0x00A0: MsgSettingsWrite,
  0x00AF: MsgSettingsWriteResp,
  0x00A4: MsgSettingsReadReq,
  0x00A5: MsgSettingsReadResp,
  0x00A2: MsgSettingsReadByIndexReq,
  0x00A7: MsgSettingsReadByIndexResp,
  0x00A6: MsgSettingsReadByIndexDone,
  0x00AE: MsgSettingsRegister,
  0x01AF: MsgSettingsRegisterResp,
}


# Copyright (C) 2015-2021 Swift Navigation Inc.
# Contact: https://support.swiftnav.com
#
# This source is subject to the license found in the file 'LICENSE' which must
# be be distributed together with this source. All other rights reserved.
#
# THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
# EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.

from .base_logger import BaseLogger
import socket


class UdpLogger(BaseLogger):
    """
    UdpLogger

    The :class:`UdpLogger` logs SBP messages over UDP.

    Parameters
    ----------
    address : string
      IP Address to send UDP packets to.
    port : int
      IP Port to send UDP packets to.
    """

    def __init__(self, address, port):
        self.handle = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.address = address
        self.port = port

    def __call__(self, msg, **metadata):
        self.call(msg, **metadata)

    def fmt_msg(self, msg, **metadata):
        return msg.pack()

    def flush(self):
        pass

    def call(self, msg, **metadata):
        self.handle.sendto(
            self.fmt_msg(msg, **metadata), (self.address, self.port))



# Copyright (C) 2017-2021 Swift Navigation Inc.
# Contact: https://support.swiftnav.com
#
# This source is subject to the license found in the file 'LICENSE' which must
# be be distributed together with this source. All other rights reserved.
#
# THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
# EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.
"""
the :mod:`sbp.client.util.fftmonitor` module contains functionality to log
and assemble spectrum analyzer messages into FFT information
"""

from collections import defaultdict
from sbp.client.drivers.network_drivers import TCPDriver
from sbp.client import Handler, Framer
from sbp.gnss import GPSTimeSec
from sbp.piksi import SBP_MSG_SPECAN_DEP
from sbp.piksi import SBP_MSG_SPECAN
import numpy as np

class GPSTimeSecCmp(GPSTimeSec):
    def __init__(self, payload=None, **kwargs):
        super(GPSTimeSecCmp, self).__init__(payload=payload, **kwargs)

    def __cmp__(self, other):
        if other is None:
            return 1
        elif self.wn < other.wn:
            return -1
        elif self.wn > other.wn:
            return 1
        else:
            if self.tow < other.tow:
                return -1
            elif self.tow > other.tow:
                return 1
            else:
                return 0

    def __eq__(self, other):
        if not isinstance(other, GPSTimeSecCmp):
            return False
        return (self.wn == other.wn) and (self.tow == other.tow)

    def __repr__(self):
        return '(wn: {0}, tow: {1})'.format(self.wn, self.tow)

    def __hash__(self):
        return (self.wn, self.tow).__hash__()


class FFTMonitor(object):
    CHANNELS = [1, 2, 3, 4]

    def __init__(self):
        self.mkvalue = lambda: {
            'frequencies': np.array([]),
            'amplitudes': np.array([])}
        self.ffts = {k: [] for k in FFTMonitor.CHANNELS}
        self.incomplete_ffts = {
            ch: defaultdict(self.mkvalue) for ch in FFTMonitor.CHANNELS}
        self.enabled = {i: None for i in FFTMonitor.CHANNELS}

    def get_frequencies(self, sbp_msg):
        ref = sbp_msg.freq_ref
        step = sbp_msg.freq_step
        n = len(sbp_msg.amplitude_value)
        return np.array([ref + step * i for i in range(n)])

    def get_amplitudes(self, sbp_msg):
        ref = sbp_msg.amplitude_ref
        unit = sbp_msg.amplitude_unit
        values = sbp_msg.amplitude_value
        return np.array([ref + unit * value for value in values])

    def capture_fft(self, sbp_msg, **metadata):

        # print "ch:%d, freq_ref: %f, freq_step: %f, amp_ref:%f, num_values:%d, wn:%d, tow:%d" \
        #         % (sbp_msg.channel_tag, sbp_msg.freq_ref, sbp_msg.freq_step, sbp_msg.amplitude_ref, len(sbp_msg.amplitude_value), sbp_msg.t.wn, sbp_msg.t.tow)

        channel = sbp_msg.channel_tag
        if not self.enabled[channel]:
            return

        frequencies = self.get_frequencies(sbp_msg)
        amplitudes = self.get_amplitudes(sbp_msg)
        timestamp = GPSTimeSecCmp(tow=sbp_msg.t.tow, wn=sbp_msg.t.wn)

        self.incomplete_ffts[channel][timestamp]['frequencies'] = np.append(
            self.incomplete_ffts[channel][timestamp]['frequencies'],
            frequencies,
            axis=0)

        self.incomplete_ffts[channel][timestamp]['amplitudes'] = np.append(
            self.incomplete_ffts[channel][timestamp]['amplitudes'],
            amplitudes,
            axis=0)

        if len(self.incomplete_ffts[channel][timestamp]['frequencies']) == 512:
            current_fft = self.incomplete_ffts[channel].pop(timestamp)
            # Clear out any partial messages
            self.incomplete_ffts[channel] = defaultdict(self.mkvalue)

            assert(len(current_fft['frequencies']) ==
                   len(current_fft['amplitudes']))
            self.ffts[channel].append(current_fft)
            #print "FFT Captured for channel %d" % channel

    def _enable(self, en, channel=None):
        if isinstance(channel, int):
            channel = [channel]
        elif channel is None:
            channel = FFTMonitor.CHANNELS
        for ch in channel:
            self.enabled[ch] = en

    def enable_channel(self, channel):
        self.clear_ffts(channel=channel)
        self._enable(True, channel=channel)

    def disable_channel(self, channel):
        self._enable(False, channel=channel)

    def num_ffts(self, channel):
        return len(self.ffts[channel])

    def get_ffts(self, channel):
        return self.ffts[channel][:]

    def clear_ffts(self, channel=None):
        if isinstance(channel, int):
            channel = [channel]
        elif channel is None:
            channel = FFTMonitor.CHANNELS

        for ch in channel:
            self.ffts[ch][:] = []
            self.incomplete_ffts[ch] = defaultdict(self.mkvalue)


def main():
    parser = argparse.ArgumentParser(
        description="Swift Navigation SBP FFT monitor example.")
    parser.add_argument(
        "-H",
        "--host",
        required=True,
        help="specify the host address.")
    parser.add_argument(
        "-p",
        "--port",
        default=55556,
        help="specify the port to use.")
    parser.add_argument(
        "-n",
        "--num-ffts",
        type=int,
        default=5,
        help="specify the number of FFTs to capture.")
    parser.add_argument(
        "-c",
        "--channel",
        type=int,
        default=1,
        help="specify the channel to monitor.")
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default="fftmonitor",
        help="specify the output filename.")
    args = parser.parse_args()

    monitor = FFTMonitor()
    ch = args.channel
    samples = args.num_ffts

    with TCPDriver(args.host, args.port) as driver:
        with Handler(Framer(driver.read, driver.write, verbose=True)) as link:
            driver.flush()
            link.add_callback(monitor.capture_fft, [SBP_MSG_SPECAN, SBP_MSG_SPECAN_DEP])

            # Capture 5 FFTs from channel 1
            monitor.enable_channel(ch)
            while monitor.num_ffts(ch) < samples:
                time.sleep(1)
            print("Captured %d ffts from channel %d" % (samples, ch))
            ffts = monitor.get_ffts(ch)
            #monitor.disable_channel(ch)

            with open('%s.pickle' % args.output, 'wb') as handle:
                pickle.dump(ffts, handle, protocol=pickle.HIGHEST_PROTOCOL)



if __name__ == "__main__":
    import argparse
    import pickle
    import time
    main()
